///////////////////////////////////////////////////////////////////////////////////////////////////
//
// PROJECT : ARIRANG family(ARIRANG-224,ARIRANG-256,ARIRANG-384,ARIRANG-512)
//
// DATE    : 2008.10.23
//
///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FILE  : ARIRANG_4S8.c
//
// NOTES : Optimized code of ARIRANG family
// 
//         Based on 8-bit platform (4 8-bit S-box version)
//
///////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <memory.h>
#include "Arirang_4S8.h"
#include "API_4S8.h"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : Init
//
// DESCRIPTION    : Initialize a hashState with the intended hash length of particular instantiation.
//
// PARAMETERS     : state - a structure that holds the hashState information
//                  hashbytelen - an integer value that indicates the length of the hash output in bytes
//
// RETURNS        : SUCCESS - on success
//                  BAD_HASHLEN - hashbytelen is invalid
//
///////////////////////////////////////////////////////////////////////////////////////////////////

HashReturn Init(hashState *state, BYTE hashbytelen)
{
	if ((hashbytelen != 28) && (hashbytelen != 32) && (hashbytelen != 48) && (hashbytelen != 64))
		return BAD_HASHLEN;

	gen_tabs();
	// Setting the Hash Length
	state->hashbytelen = hashbytelen;
	
	// Setting the Counter Values
	memset(state->counter,0,16);
	
	// Initialize the Data Length
	memset(state->count,0,16);

	// Setting the Initial Hash Value
	if(state->hashbytelen == 28){
		state->workingvar[ 0] = 0xcb; state->workingvar[ 1] = 0xbb; state->workingvar[ 2] = 0x9d; state->workingvar[ 3] = 0x5d;
		state->workingvar[ 4] = 0x62; state->workingvar[ 5] = 0x9a; state->workingvar[ 6] = 0x29; state->workingvar[ 7] = 0x2a;
		state->workingvar[ 8] = 0x91; state->workingvar[ 9] = 0x59; state->workingvar[10] = 0x01; state->workingvar[11] = 0x5a;
		state->workingvar[12] = 0x15; state->workingvar[13] = 0x2f; state->workingvar[14] = 0xec; state->workingvar[15] = 0xd8;
		state->workingvar[16] = 0x67; state->workingvar[17] = 0x33; state->workingvar[18] = 0x26; state->workingvar[19] = 0x67;
		state->workingvar[20] = 0x8e; state->workingvar[21] = 0xb4; state->workingvar[22] = 0x4a; state->workingvar[23] = 0x87;
		state->workingvar[24] = 0xdb; state->workingvar[25] = 0x0c; state->workingvar[26] = 0x2e; state->workingvar[27] = 0x0d;
		state->workingvar[28] = 0x47; state->workingvar[29] = 0xb5; state->workingvar[30] = 0x48; state->workingvar[31] = 0x1d;
		state->blocklen=ARIRANG256_BLOCK_LEN;
	}
		
	else if(state->hashbytelen == 32){
		state->workingvar[ 0] = 0x6a; state->workingvar[ 1] = 0x09; state->workingvar[ 2] = 0xe6; state->workingvar[ 3] = 0x67;
		state->workingvar[ 4] = 0xbb; state->workingvar[ 5] = 0x67; state->workingvar[ 6] = 0xae; state->workingvar[ 7] = 0x85;
		state->workingvar[ 8] = 0x3c; state->workingvar[ 9] = 0x6e; state->workingvar[10] = 0xf3; state->workingvar[11] = 0x72;
		state->workingvar[12] = 0xa5; state->workingvar[13] = 0x4f; state->workingvar[14] = 0xf5; state->workingvar[15] = 0x3a;
		state->workingvar[16] = 0x51; state->workingvar[17] = 0x0e; state->workingvar[18] = 0x52; state->workingvar[19] = 0x7f;
		state->workingvar[20] = 0x9b; state->workingvar[21] = 0x05; state->workingvar[22] = 0x68; state->workingvar[23] = 0x8c;
		state->workingvar[24] = 0x1f; state->workingvar[25] = 0x83; state->workingvar[26] = 0xd9; state->workingvar[27] = 0xab;
		state->workingvar[28] = 0x5b; state->workingvar[29] = 0xe0; state->workingvar[30] = 0xcd; state->workingvar[31] = 0x19;
		state->blocklen=ARIRANG256_BLOCK_LEN;
	
	}
	else if(hashbytelen == 48){	                                                                                       
		state->workingvar[ 0]=0xcb; state->workingvar[ 1]=0xbb; state->workingvar[ 2]=0x9d; state->workingvar[ 3]=0x5d;
		state->workingvar[ 4]=0xc1; state->workingvar[ 5]=0x05; state->workingvar[ 6]=0x9e; state->workingvar[ 7]=0xd8;
		state->workingvar[ 8]=0x62; state->workingvar[ 9]=0x9a; state->workingvar[10]=0x29; state->workingvar[11]=0x2a;
		state->workingvar[12]=0x36; state->workingvar[13]=0x7c; state->workingvar[14]=0xd5; state->workingvar[15]=0x07;
		state->workingvar[16]=0x91; state->workingvar[17]=0x59; state->workingvar[18]=0x01; state->workingvar[19]=0x5a;
		state->workingvar[20]=0x30; state->workingvar[21]=0x70; state->workingvar[22]=0xdd; state->workingvar[23]=0x17;
		state->workingvar[24]=0x15; state->workingvar[25]=0x2f; state->workingvar[26]=0xec; state->workingvar[27]=0xd8;
		state->workingvar[28]=0xf7; state->workingvar[29]=0x0e; state->workingvar[30]=0x59; state->workingvar[31]=0x39;
		state->workingvar[32]=0x67; state->workingvar[33]=0x33; state->workingvar[34]=0x26; state->workingvar[35]=0x67;
		state->workingvar[36]=0xff; state->workingvar[37]=0xc0; state->workingvar[38]=0x0b; state->workingvar[39]=0x31;
		state->workingvar[40]=0x8e; state->workingvar[41]=0xb4; state->workingvar[42]=0x4a; state->workingvar[43]=0x87;
		state->workingvar[44]=0x68; state->workingvar[45]=0x58; state->workingvar[46]=0x15; state->workingvar[47]=0x11;
		state->workingvar[48]=0xdb; state->workingvar[49]=0x0c; state->workingvar[50]=0x2e; state->workingvar[51]=0x0d;
		state->workingvar[52]=0x64; state->workingvar[53]=0xf9; state->workingvar[54]=0x8f; state->workingvar[55]=0xa7;
		state->workingvar[56]=0x47; state->workingvar[57]=0xb5; state->workingvar[58]=0x48; state->workingvar[59]=0x1d;
		state->workingvar[60]=0xbe; state->workingvar[61]=0xfa; state->workingvar[62]=0x4f; state->workingvar[63]=0xa4;
		state->blocklen=ARIRANG512_BLOCK_LEN;	                                                                         
	}													                                                                                       
	else if(hashbytelen == 64){									                                                                     
		state->workingvar[ 0]=0x6a; state->workingvar[ 1]=0x09; state->workingvar[ 2]=0xe6; state->workingvar[ 3]=0x67;
		state->workingvar[ 4]=0xf3; state->workingvar[ 5]=0xbc; state->workingvar[ 6]=0xc9; state->workingvar[ 7]=0x08;
		state->workingvar[ 8]=0xbb; state->workingvar[ 9]=0x67; state->workingvar[10]=0xae; state->workingvar[11]=0x85;
		state->workingvar[12]=0x84; state->workingvar[13]=0xca; state->workingvar[14]=0xa7; state->workingvar[15]=0x3b;
		state->workingvar[16]=0x3c; state->workingvar[17]=0x6e; state->workingvar[18]=0xf3; state->workingvar[19]=0x72;
		state->workingvar[20]=0xfe; state->workingvar[21]=0x94; state->workingvar[22]=0xf8; state->workingvar[23]=0x2b;
		state->workingvar[24]=0xa5; state->workingvar[25]=0x4f; state->workingvar[26]=0xf5; state->workingvar[27]=0x3a;
		state->workingvar[28]=0x5f; state->workingvar[29]=0x1d; state->workingvar[30]=0x36; state->workingvar[31]=0xf1;
		state->workingvar[32]=0x51; state->workingvar[33]=0x0e; state->workingvar[34]=0x52; state->workingvar[35]=0x7f;
		state->workingvar[36]=0xad; state->workingvar[37]=0xe6; state->workingvar[38]=0x82; state->workingvar[39]=0xd1;
		state->workingvar[40]=0x9b; state->workingvar[41]=0x05; state->workingvar[42]=0x68; state->workingvar[43]=0x8c;
		state->workingvar[44]=0x2b; state->workingvar[45]=0x3e; state->workingvar[46]=0x6c; state->workingvar[47]=0x1f;
		state->workingvar[48]=0x1f; state->workingvar[49]=0x83; state->workingvar[50]=0xd9; state->workingvar[51]=0xab;
		state->workingvar[52]=0xfb; state->workingvar[53]=0x41; state->workingvar[54]=0xbd; state->workingvar[55]=0x6b;
		state->workingvar[56]=0x5b; state->workingvar[57]=0xe0; state->workingvar[58]=0xcd; state->workingvar[59]=0x19;
		state->workingvar[60]=0x13; state->workingvar[61]=0x7e; state->workingvar[62]=0x21; state->workingvar[63]=0x79;
		state->blocklen=ARIRANG512_BLOCK_LEN;		                                                                       
	}                                                                                                                

	
	return SUCCESS;

}


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : Update
//
// DESCRIPTION    : Process the supplied data.
//
// PARAMETERS     : state - a structure that holds the hashState information
//                  data - the data to be hashed
//                  Databitlen - the length, in bits, of the data to be hashed
//
// RETURNS        : SUCCESS - on success
//
///////////////////////////////////////////////////////////////////////////////////////////////////

HashReturn Update(hashState *state, const BitSequence *data, BYTE* databitlen)
{
	BYTE RemainedLen, PartLen, temp;
	BYTE databytelen;

	
	databytelen = ( databitlen[0]+7 >> 3) + (databitlen[1] << 5 );
	
	
	// Compute the number of hashed bytes mod ARIRANG_BLOCK_LEN
	RemainedLen = ((state->count[0] >> 3) + (state->count[1] << 5)) % state->blocklen;

	// Compute the number of bytes that can be filled up
	PartLen = state->blocklen - RemainedLen;

	// Update count (number of toatl data bits)
	temp = state->count[0] + (databytelen << 3);
	if(temp < state->count[0] ){
		  state->count[ 1]++; if(state->count[ 1] != 0x00);
	else {state->count[ 2]++; if(state->count[ 2] != 0x00); 
	else {state->count[ 3]++; if(state->count[ 3] != 0x00);
	else {state->count[ 4]++; if(state->count[ 4] != 0x00);
	else {state->count[ 5]++; if(state->count[ 5] != 0x00);
	else {state->count[ 6]++; if(state->count[ 6] != 0x00);
	else {state->count[ 7]++; if(state->count[ 7] != 0x00); 
	else {state->count[ 8]++; if(state->count[ 8] != 0x00);
	else {state->count[ 9]++; if(state->count[ 9] != 0x00);
	else {state->count[10]++; if(state->count[10] != 0x00);
	else {state->count[11]++; if(state->count[11] != 0x00);
	else {state->count[12]++; if(state->count[12] != 0x00); 
	else {state->count[13]++; if(state->count[13] != 0x00);
	else {state->count[14]++; if(state->count[14] != 0x00);
	else {state->count[15]++;}}}}}}}}}}}}}}}
	state->count[0]=temp;

	temp = state->count[1] + databitlen[1];
	if(temp < state->count[1] ){
		  state->count[ 2]++; if(state->count[ 2] != 0x00);
	else {state->count[ 3]++; if(state->count[ 3] != 0x00); 
	else {state->count[ 4]++; if(state->count[ 4] != 0x00);
	else {state->count[ 5]++; if(state->count[ 5] != 0x00);
	else {state->count[ 6]++; if(state->count[ 6] != 0x00);
	else {state->count[ 7]++; if(state->count[ 7] != 0x00); 
	else {state->count[ 8]++; if(state->count[ 8] != 0x00);
	else {state->count[ 9]++; if(state->count[ 9] != 0x00);
	else {state->count[10]++; if(state->count[10] != 0x00);
	else {state->count[11]++; if(state->count[11] != 0x00);
	else {state->count[12]++; if(state->count[12] != 0x00); 
	else {state->count[13]++; if(state->count[13] != 0x00);
	else {state->count[14]++; if(state->count[14] != 0x00);
	else {state->count[15]++;}}}}}}}}}}}}}}
	state->count[1]=temp;
	
	if (((databytelen >= PartLen) && (databitlen[0] == 0)) || ((databytelen == PartLen) && (state->remainderbit == 0)) ) {
		memcpy(state->block + RemainedLen, data, PartLen);
		if(state->hashbytelen < 33) Arirang_Compression256(state);
		else Arirang_Compression512(state);

		data += PartLen;
		databytelen -= PartLen;
		RemainedLen = 0;

		while( (databytelen > state->blocklen) || ((databytelen == state->blocklen) && (state->remainderbit == 0)) ) {
			memcpy((BYTE *)state->block, data, state->blocklen);
			if(state->hashbytelen < 33) Arirang_Compression256(state);
			else Arirang_Compression512(state);

			data += state->blocklen;
			databytelen -= state->blocklen;
		}
	}

	//	Buffer remaining input
	memcpy((BYTE *)state->block + RemainedLen, data, databytelen);

	return SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : Final
//
// DESCRIPTION    : Perform any post processing and ouput filtering required and return the final 
//                  hash value.
//
// PARAMETERS     : state - a structure that holds the hashState information
//                  hashval - the storage for the final hash value to be returned
//
// RETURNS        : SUCCESS - on success
//
///////////////////////////////////////////////////////////////////////////////////////////////////

HashReturn Final(hashState *state, BYTE *hashval)
{
	BYTE i, dwIndex, count[16];
	BYTE temp = (state->blocklen >> 3);

	// Padding the message
	if(state->remainderbit){
		count[ 0] = state->count[ 0] + state->remainderbit - 8; count[ 1] = state->count[ 1]; count[ 2] = state->count[ 2]; count[ 3] = state->count[ 3];
		count[ 4] = state->count[ 4]; count[ 5] = state->count[ 5]; count[ 6] = state->count[ 6]; count[ 7] = state->count[ 7];
		count[ 8] = state->count[ 8]; count[ 9] = state->count[ 9]; count[10] = state->count[10]; count[11] = state->count[11];
		count[12] = state->count[12]; count[13] = state->count[13]; count[14] = state->count[14]; count[15] = state->count[15];

		dwIndex = ((count[0] >> 3) + (count[1] << 5)) % state->blocklen;
		state->block[dwIndex] &= 0xff-(1<<(8-state->remainderbit))+1;		
		state->block[dwIndex++] ^= 0x80>>(state->remainderbit);
	}
	else{
		count[ 0] = state->count[ 0]; count[ 1] = state->count[ 1]; count[ 2] = state->count[ 2]; count[ 3] = state->count[ 3];
		count[ 4] = state->count[ 4]; count[ 5] = state->count[ 5]; count[ 6] = state->count[ 6]; count[ 7] = state->count[ 7];
		count[ 8] = state->count[ 8]; count[ 9] = state->count[ 9]; count[10] = state->count[10]; count[11] = state->count[11];
		count[12] = state->count[12]; count[13] = state->count[13]; count[14] = state->count[14]; count[15] = state->count[15];
		dwIndex = ((count[0] >> 3) + (count[1] << 5)) % state->blocklen;		
		state->block[dwIndex++] = 0x80;	
	}
		
	if (dwIndex > (state->blocklen - temp)){
		memset((BYTE *)state->block + dwIndex, 0, (BYTE)(state->blocklen - dwIndex));
		if(state->hashbytelen < 33) Arirang_Compression256(state);
		else Arirang_Compression512(state);

		memset((BYTE *)state->block, 0, (BYTE)state->blocklen - temp);
	}
	else
		memset((BYTE *)state->block + dwIndex, 0, (BYTE)(state->blocklen - dwIndex - temp));

	// Fixed counter value for the last message block
	if(state->hashbytelen > 33){
		state->block[state->blocklen-16] = count[15]; state->block[state->blocklen-15] = count[14];	state->block[state->blocklen-14] = count[13]; state->block[state->blocklen-13] = count[12];
		state->block[state->blocklen-12] = count[11]; state->block[state->blocklen-11] = count[10];	state->block[state->blocklen-10] = count[ 9]; state->block[state->blocklen- 9] = count[ 8];
		state->block[state->blocklen- 8] = count[ 7]; state->block[state->blocklen- 7] = count[ 6];	state->block[state->blocklen- 6] = count[ 5]; state->block[state->blocklen- 5] = count[ 4];
		state->block[state->blocklen- 4] = count[ 3]; state->block[state->blocklen- 3] = count[ 2];	state->block[state->blocklen- 2] = count[ 1]; state->block[state->blocklen- 1] = count[ 0];
		state->counter[15]=0xb7; state->counter[14]=0xe1; state->counter[13]=0x51; state->counter[12]=0x62;	
		state->counter[11]=0x8a; state->counter[10]=0xed; state->counter[ 9]=0x2a; state->counter[ 8]=0x6a;
		state->counter[ 7]=0xbf; state->counter[ 6]=0x71; state->counter[ 5]=0x58; state->counter[ 4]=0x80;	
		state->counter[ 3]=0x9c; state->counter[ 2]=0xf4; state->counter[ 1]=0xf3; state->counter[ 0]=0xc7;
	}
	else{
		state->block[state->blocklen-8] = count[7];	state->block[state->blocklen-7] = count[6];	state->block[state->blocklen-6] = count[5];	state->block[state->blocklen-5] = count[4];
		state->block[state->blocklen-4] = count[3];	state->block[state->blocklen-3] = count[2];	state->block[state->blocklen-2] = count[1];	state->block[state->blocklen-1] = count[0];
		state->counter[7]=0xb7; state->counter[6]=0xe1; state->counter[5]=0x51; state->counter[4]=0x62;	
		state->counter[3]=0x8a; state->counter[2]=0xed; state->counter[1]=0x2a; state->counter[0]=0x6a;	
	}
	
	if(state->hashbytelen < 33) Arirang_Compression256(state);
	else Arirang_Compression512(state);

	for (i = 0; i < state->hashbytelen ; i++)	hashval[i]=state->workingvar[i]; 

	return SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : Hash
//
// DESCRIPTION    : Hash the supplied data and provide the resulting hash value. 
//                  
//
// PARAMETERS     : state      - a structure that holds the hashState information
//                  data       - the data to be hashed
//                  databitlen - the length, in bits, of the data to be hashed
//                  hashval    - the storage for the final hash value to be returned
//
// RETURNS        : SUCCESS - on success
//
///////////////////////////////////////////////////////////////////////////////////////////////////
HashReturn Hash(BYTE hashbytelen, const BitSequence *data, BYTE *databitlen, BitSequence *hashval){

	hashState State;
	BitSequence *UpdatedData;	
	BYTE UpdatedDataLengthbit[2]={0,1};
	BYTE i[15]={0,};
	
	UpdatedData=data;
	State.remainderbit = databitlen[0] & 7;

	Init(&State, hashbytelen);
	while((i[ 0] < databitlen[ 1]) || (i[ 1] < databitlen[ 2]) || (i[ 2] < databitlen[ 3]) || (i[ 3] < databitlen[ 4]) ||
		  (i[ 4] < databitlen[ 5]) || (i[ 5] < databitlen[ 6]) || (i[ 6] < databitlen[ 7]) || (i[ 7] < databitlen[ 8]) ||
		  (i[ 8] < databitlen[ 9]) || (i[ 9] < databitlen[10]) || (i[10] < databitlen[11]) || (i[11] < databitlen[12]) ||
		  (i[12] < databitlen[13]) || (i[13] < databitlen[14]) || (i[14] < databitlen[15]) ){
		Update(&State, UpdatedData, UpdatedDataLengthbit);
		UpdatedData+=0x20;
		i[0]++;  if((i[0] != 0x00));
		else {i[ 1]++;if(i[ 1] != 0x00);
		else {i[ 2]++;if(i[ 2] != 0x00);
		else {i[ 3]++;if(i[ 3] != 0x00);
		else {i[ 4]++;if(i[ 4] != 0x00);
		else {i[ 5]++;if(i[ 5] != 0x00);
		else {i[ 6]++;if(i[ 6] != 0x00);
		else {i[ 7]++;if(i[ 7] != 0x00);
		else {i[ 8]++;if(i[ 8] != 0x00);
		else {i[ 9]++;if(i[ 9] != 0x00);
		else {i[10]++;if(i[10] != 0x00);
		else {i[11]++;if(i[11] != 0x00);
		else {i[12]++;if(i[12] != 0x00);
		else {i[13]++;if(i[13] != 0x00);
		else {i[14]++;if(i[14] != 0x00);
		else {i[15]++;}}}}}}}}}}}}}}}
	}
	UpdatedDataLengthbit[1] = 0;
	UpdatedDataLengthbit[0] = databitlen[0];
	Update(&State, UpdatedData, UpdatedDataLengthbit);

	Final(&State, hashval);
	return SUCCESS;
	
}