///////////////////////////////////////////////////////////////////////////////////////////////////
//
// PROJECT : ARIRANG family(ARIRANG-224,ARIRANG-256,ARIRANG-384,ARIRANG-512)
//
// DATE    : 2008.10.23
//
///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FILE  : ARIRANG_4S8.c
//
// NOTES : Optimized code of ARIRANG family
// 
//         Based on 8-bit platform (4 8-bit S-box version)
//
///////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>

#include <memory.h>

#include "Arirang_4S8.h"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Macro

#undef LL
#undef RR
#define LL(x,n)		(x << n)
#define RR(x,n)		(x >> n)

#undef SS0
#undef SS1
#undef SS2
#undef SS3
#undef SSS0
#undef SSS1
#undef SSS2
#undef SSS3
#undef SSS4
#undef SSS5
#undef SSS6
#undef SSS7

#define SS0(x,y,n1,n2)	x[0] ^= RR(y[1],n1)^LL(y[0],n2); x[1] ^= RR(y[2],n1)^LL(y[1],n2); x[2] ^= RR(y[3],n1)^LL(y[2],n2); x[3] ^= RR(y[0],n1)^LL(y[3],n2);
#define SS1(x,y,n1,n2)	x[3] ^= RR(y[1],n1)^LL(y[0],n2); x[0] ^= RR(y[2],n1)^LL(y[1],n2); x[1] ^= RR(y[3],n1)^LL(y[2],n2); x[2] ^= RR(y[0],n1)^LL(y[3],n2);
#define SS2(x,y,n1,n2)	x[2] ^= RR(y[1],n1)^LL(y[0],n2); x[3] ^= RR(y[2],n1)^LL(y[1],n2); x[0] ^= RR(y[3],n1)^LL(y[2],n2); x[1] ^= RR(y[0],n1)^LL(y[3],n2);
#define SS3(x,y,n1,n2)	x[1] ^= RR(y[1],n1)^LL(y[0],n2); x[2] ^= RR(y[2],n1)^LL(y[1],n2); x[3] ^= RR(y[3],n1)^LL(y[2],n2); x[0] ^= RR(y[0],n1)^LL(y[3],n2);

#define SSS0(x,y,n1,n2)	x[0] ^= RR(y[1],n1)^LL(y[0],n2); x[1] ^= RR(y[2],n1)^LL(y[1],n2); x[2] ^= RR(y[3],n1)^LL(y[2],n2); x[3] ^= RR(y[4],n1)^LL(y[3],n2); x[4] ^= RR(y[5],n1)^LL(y[4],n2); x[5] ^= RR(y[6],n1)^LL(y[5],n2); x[6] ^= RR(y[7],n1)^LL(y[6],n2); x[7] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS1(x,y,n1,n2)	x[7] ^= RR(y[1],n1)^LL(y[0],n2); x[0] ^= RR(y[2],n1)^LL(y[1],n2); x[1] ^= RR(y[3],n1)^LL(y[2],n2); x[2] ^= RR(y[4],n1)^LL(y[3],n2); x[3] ^= RR(y[5],n1)^LL(y[4],n2); x[4] ^= RR(y[6],n1)^LL(y[5],n2); x[5] ^= RR(y[7],n1)^LL(y[6],n2); x[6] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS2(x,y,n1,n2)	x[6] ^= RR(y[1],n1)^LL(y[0],n2); x[7] ^= RR(y[2],n1)^LL(y[1],n2); x[0] ^= RR(y[3],n1)^LL(y[2],n2); x[1] ^= RR(y[4],n1)^LL(y[3],n2); x[2] ^= RR(y[5],n1)^LL(y[4],n2); x[3] ^= RR(y[6],n1)^LL(y[5],n2); x[4] ^= RR(y[7],n1)^LL(y[6],n2); x[5] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS3(x,y,n1,n2)	x[5] ^= RR(y[1],n1)^LL(y[0],n2); x[6] ^= RR(y[2],n1)^LL(y[1],n2); x[7] ^= RR(y[3],n1)^LL(y[2],n2); x[0] ^= RR(y[4],n1)^LL(y[3],n2); x[1] ^= RR(y[5],n1)^LL(y[4],n2); x[2] ^= RR(y[6],n1)^LL(y[5],n2); x[3] ^= RR(y[7],n1)^LL(y[6],n2); x[4] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS4(x,y,n1,n2)	x[4] ^= RR(y[1],n1)^LL(y[0],n2); x[5] ^= RR(y[2],n1)^LL(y[1],n2); x[6] ^= RR(y[3],n1)^LL(y[2],n2); x[7] ^= RR(y[4],n1)^LL(y[3],n2); x[0] ^= RR(y[5],n1)^LL(y[4],n2); x[1] ^= RR(y[6],n1)^LL(y[5],n2); x[2] ^= RR(y[7],n1)^LL(y[6],n2); x[3] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS5(x,y,n1,n2)	x[3] ^= RR(y[1],n1)^LL(y[0],n2); x[4] ^= RR(y[2],n1)^LL(y[1],n2); x[5] ^= RR(y[3],n1)^LL(y[2],n2); x[6] ^= RR(y[4],n1)^LL(y[3],n2); x[7] ^= RR(y[5],n1)^LL(y[4],n2); x[0] ^= RR(y[6],n1)^LL(y[5],n2); x[1] ^= RR(y[7],n1)^LL(y[6],n2); x[2] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS6(x,y,n1,n2)	x[2] ^= RR(y[1],n1)^LL(y[0],n2); x[3] ^= RR(y[2],n1)^LL(y[1],n2); x[4] ^= RR(y[3],n1)^LL(y[2],n2); x[5] ^= RR(y[4],n1)^LL(y[3],n2); x[6] ^= RR(y[5],n1)^LL(y[4],n2); x[7] ^= RR(y[6],n1)^LL(y[5],n2); x[0] ^= RR(y[7],n1)^LL(y[6],n2); x[1] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS7(x,y,n1,n2)	x[1] ^= RR(y[1],n1)^LL(y[0],n2); x[2] ^= RR(y[2],n1)^LL(y[1],n2); x[3] ^= RR(y[3],n1)^LL(y[2],n2); x[4] ^= RR(y[4],n1)^LL(y[3],n2); x[5] ^= RR(y[5],n1)^LL(y[4],n2); x[6] ^= RR(y[6],n1)^LL(y[5],n2); x[7] ^= RR(y[7],n1)^LL(y[6],n2); x[0] ^= RR(y[0],n1)^LL(y[7],n2);

#define f_mult2(x) (x&0x80 ? (0x1b^(x<<1)) : (x<<1))
///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Constant
static const BYTE K256[64] = {
	0x51, 0x7c, 0xc1, 0xb7, 0x76, 0x51, 0x7c, 0xc1, 0xbd, 0x76, 0x51, 0x7c, 0x2d, 0xbd, 0x76, 0x51,
	0x27, 0x2d, 0xbd, 0x76, 0xcb, 0x27, 0x2d, 0xbd, 0x90, 0xcb, 0x27, 0x2d, 0x0a, 0x90, 0xcb, 0x27,
	0xec, 0x0a, 0x90, 0xcb, 0x5b, 0xec, 0x0a, 0x90, 0x9a, 0x5b, 0xec, 0x0a, 0xe6, 0x9a, 0x5b, 0xec,
	0xb7, 0xe6, 0x9a, 0x5b, 0xc1, 0xb7, 0xe6, 0x9a, 0x7c, 0xc1, 0xb7, 0xe6, 0x51, 0x7c, 0xc1, 0xb7
};

static const BYTE K512[128] = {
	0x51, 0x7c, 0xc1, 0xb7, 0x27, 0x22, 0x0a, 0x94, 0x2d, 0xb6, 0x51, 0x7c, 0xc1, 0xb7, 0x27, 0x22, 0xe6, 0x95, 0x2d, 0xb6, 0x51, 0x7c, 0xc1, 0xb7, 0x90, 0xcb, 0xe6, 0x95, 0x2d, 0xb6, 0x51, 0x7c,
	0x7c, 0xca, 0x90, 0xcb, 0xe6, 0x95, 0x2d, 0xb6, 0xcb, 0x23, 0x7c, 0xca, 0x90, 0xcb, 0xe6, 0x95, 0x76, 0x5e, 0xcb, 0x23, 0x7c, 0xca, 0x90, 0xcb, 0xec, 0x01, 0x76, 0x5e, 0xcb, 0x23, 0x7c, 0xca,
	0xb7, 0xe9, 0xec, 0x01, 0x76, 0x5e, 0xcb, 0x23, 0xbd, 0x7d, 0xb7, 0xe9, 0xec, 0x01, 0x76, 0x5e, 0x9a, 0x5f, 0xbd, 0x7d, 0xb7, 0xe9, 0xec, 0x01, 0x5b, 0xe8, 0x9a, 0x5f, 0xbd, 0x7d, 0xb7, 0xe9,
	0x0a, 0x94, 0x5b, 0xe8, 0x9a, 0x5f, 0xbd, 0x7d, 0x27, 0x22, 0x0a, 0x94, 0x5b, 0xe8, 0x9a, 0x5f, 0xc1, 0xb7, 0x27, 0x22, 0x0a, 0x94, 0x5b, 0xe8, 0x51, 0x7c, 0xc1, 0xb7, 0x27, 0x22, 0x0a, 0x94
};



BYTE	sbx1[256]; // S-box: S[i]
BYTE	sbx2[256]; // 2*S[i] GF(256)
BYTE	sbx4[256]; // 4*S[i] GF(256)
BYTE	sbx8[256]; // 8*S[i] GF(256)


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : gen_tabs
//
// DESCRIPTION    : Generate the s-box table and field multiplication table
//
// PARAMETERS     : void
//                  
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////
void gen_tabs(void)
{   

	#ifndef __GEN_TABS__
	#define __GEN_TABS__ 100
	BYTE i; 	
    BYTE p, q;
	
 
	/* log and power tables for GF(2**8) finite field with  */
    /* 0x011b as modular polynomial - the simplest primitive */
    /* root is 0x03, used here to generate the tables       */

	i = 0;p = 1;
    while(1)
    {
        sbx4[i] = (BYTE)p; //pow_tab ==> sbx4
		sbx8[p] = (BYTE)i; //log_tab ==> sbx8 
        p = p ^ (p << 1) ^ (p & 0x80 ? 0x01b : 0);
		if(i == 255) break;
		i++;
    }

    /* note that the affine byte transformation matrix in   */
    /* ARIRANG specification is in big endian format with  */
    /* bit 0 as the most significant bit. In the remainder  */
    /* of the specification the bits are numbered from the  */
    /* least significant end of a byte.                     */

	i = 0;
	while(1)    
    {   
        p = (i ? sbx4[255 - sbx8[i]] : 0); q = p; 
        q = (q >> 7) | (q << 1); p ^= q; 
        q = (q >> 7) | (q << 1); p ^= q; 
        q = (q >> 7) | (q << 1); p ^= q; 
        q = (q >> 7) | (q << 1); p ^= q ^ 0x63; 
        sbx1[i] = p;
		if(i == 0xff) break;
		i++;
    }
	i = 0;
	while(1)    
    {   
		sbx2[i] = f_mult2(sbx1[i]);
		sbx4[i] = f_mult2(sbx2[i]);
		sbx8[i] = f_mult2(sbx4[i]);
		if(i == 0xff) break;
		i++;
    }
	#endif
}
///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : step256
//
// DESCRIPTION    : Step function of ARIRANG-224 and ARIRANG-256
//
// PARAMETERS     : A,B,C,D,E,F,G,H  - working variables
//                  M1,M2 - message block
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////

void Step256(BYTE* A, BYTE* B, BYTE* C, BYTE* D, BYTE* E, BYTE* F, BYTE* G, BYTE* H, BYTE* M1, BYTE* M2){
	BYTE temp1[4], temp2[4];
	A[0] ^= M1[0]; A[1] ^= M1[1]; A[2] ^= M1[2]; A[3] ^= M1[3];							
	E[0] ^= M2[0]; E[1] ^= M2[1]; E[2] ^= M2[2]; E[3] ^= M2[3];							
	temp1[3] = sbx2[A[3]] ^(sbx2[A[2]] ^ sbx1[A[2]])^ sbx1[A[1]] ^ sbx1[A[0]];		       			
	temp1[2] = sbx1[A[3]] ^ sbx2[A[2]] ^(sbx2[A[1]] ^ sbx1[A[1]])^ sbx1[A[0]];						
	temp1[1] = sbx1[A[3]] ^ sbx1[A[2]] ^ sbx2[A[1]] ^(sbx2[A[0]] ^ sbx1[A[0]]);						
	temp1[0] =(sbx2[A[3]] ^ sbx1[A[3]])^ sbx1[A[2]] ^ sbx1[A[1]] ^ sbx2[A[0]];						
	temp2[3] = sbx2[E[3]] ^(sbx2[E[2]] ^ sbx1[E[2]])^ sbx1[E[1]] ^ sbx1[E[0]];		       			
	temp2[2] = sbx1[E[3]] ^ sbx2[E[2]] ^(sbx2[E[1]] ^ sbx1[E[1]])^ sbx1[E[0]];						
	temp2[1] = sbx1[E[3]] ^ sbx1[E[2]] ^ sbx2[E[1]] ^(sbx2[E[0]] ^ sbx1[E[0]]);						
	temp2[0] =(sbx2[E[3]] ^ sbx1[E[3]])^ sbx1[E[2]] ^ sbx1[E[1]] ^ sbx2[E[0]];						
	B[0] ^= temp1[0];	B[1] ^= temp1[1];	B[2] ^= temp1[2];	B[3] ^= temp1[3];		
	SS1(C,temp1,3,5);																	
	SS2(D,temp1,1,7);																	
	F[0] ^= temp2[0];	F[1] ^= temp2[1];	F[2] ^= temp2[2];	F[3] ^= temp2[3];		
	SS3(G,temp2,3,5);																	
	SS0(H,temp2,1,7);																	
}	
																					
///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : step512
//
// DESCRIPTION    : Step function of ARIRANG-384 and Arirang-512
//
// PARAMETERS     : A,B,C,D,E,F,G,H  - working variables
//                  M1,M2 - message block
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////
void Step512(BYTE* A, BYTE* B, BYTE* C, BYTE* D, BYTE* E, BYTE* F, BYTE* G, BYTE* H, BYTE* M1, BYTE* M2){
	BYTE temp1[8], temp2[8];
	A[0] ^= M1[0]; A[1] ^= M1[1]; A[2] ^= M1[2]; A[3] ^= M1[3];															
	A[4] ^= M1[4]; A[5] ^= M1[5]; A[6] ^= M1[6]; A[7] ^= M1[7];															
	E[0] ^= M2[0]; E[1] ^= M2[1]; E[2] ^= M2[2]; E[3] ^= M2[3];															
	E[4] ^= M2[4]; E[5] ^= M2[5]; E[6] ^= M2[6]; E[7] ^= M2[7];															
	temp1[7] = sbx1[A[7]] ^ sbx2[A[6]] ^(sbx2[A[5]] ^ sbx8[A[5]])^(sbx1[A[4]] ^ sbx8[A[4]])^ sbx8[A[3]] ^ sbx1[A[2]] ^ sbx4[A[1]] ^ sbx1[A[0]];	
	temp1[6] = sbx1[A[7]] ^ sbx1[A[6]] ^ sbx2[A[5]] ^(sbx2[A[4]] ^ sbx8[A[4]])^(sbx1[A[3]] ^ sbx8[A[3]])^ sbx8[A[2]] ^ sbx1[A[1]] ^ sbx4[A[0]];	
	temp1[5] = sbx4[A[7]] ^ sbx1[A[6]] ^ sbx1[A[5]] ^ sbx2[A[4]] ^(sbx2[A[3]] ^ sbx8[A[3]])^(sbx1[A[2]] ^ sbx8[A[2]])^ sbx8[A[1]] ^ sbx1[A[0]];	
	temp1[4] = sbx1[A[7]] ^ sbx4[A[6]] ^ sbx1[A[5]] ^ sbx1[A[4]] ^ sbx2[A[3]] ^(sbx2[A[2]] ^ sbx8[A[2]])^(sbx1[A[1]] ^ sbx8[A[1]])^ sbx8[A[0]];	
	temp1[3] = sbx8[A[7]] ^ sbx1[A[6]] ^ sbx4[A[5]] ^ sbx1[A[4]] ^ sbx1[A[3]] ^ sbx2[A[2]] ^(sbx2[A[1]] ^ sbx8[A[1]])^(sbx1[A[0]] ^ sbx8[A[0]]);	
	temp1[2] =(sbx1[A[7]] ^ sbx8[A[7]])^ sbx8[A[6]] ^ sbx1[A[5]] ^ sbx4[A[4]] ^ sbx1[A[3]] ^ sbx1[A[2]] ^ sbx2[A[1]] ^(sbx2[A[0]] ^ sbx8[A[0]]);	
	temp1[1] =(sbx2[A[7]] ^ sbx8[A[7]])^(sbx1[A[6]] ^ sbx8[A[6]])^ sbx8[A[5]] ^ sbx1[A[4]] ^ sbx4[A[3]] ^ sbx1[A[2]] ^ sbx1[A[1]] ^ sbx2[A[0]];	
	temp1[0] = sbx2[A[7]] ^(sbx2[A[6]] ^ sbx8[A[6]])^(sbx1[A[5]] ^ sbx8[A[5]])^ sbx8[A[4]] ^ sbx1[A[3]] ^ sbx4[A[2]] ^ sbx1[A[1]] ^ sbx1[A[0]];	
	temp2[7] = sbx1[E[7]] ^ sbx2[E[6]] ^(sbx2[E[5]] ^ sbx8[E[5]])^(sbx1[E[4]] ^ sbx8[E[4]])^ sbx8[E[3]] ^ sbx1[E[2]] ^ sbx4[E[1]] ^ sbx1[E[0]];	
	temp2[6] = sbx1[E[7]] ^ sbx1[E[6]] ^ sbx2[E[5]] ^(sbx2[E[4]] ^ sbx8[E[4]])^(sbx1[E[3]] ^ sbx8[E[3]])^ sbx8[E[2]] ^ sbx1[E[1]] ^ sbx4[E[0]];	
	temp2[5] = sbx4[E[7]] ^ sbx1[E[6]] ^ sbx1[E[5]] ^ sbx2[E[4]] ^(sbx2[E[3]] ^ sbx8[E[3]])^(sbx1[E[2]] ^ sbx8[E[2]])^ sbx8[E[1]] ^ sbx1[E[0]];	
	temp2[4] = sbx1[E[7]] ^ sbx4[E[6]] ^ sbx1[E[5]] ^ sbx1[E[4]] ^ sbx2[E[3]] ^(sbx2[E[2]] ^ sbx8[E[2]])^(sbx1[E[1]] ^ sbx8[E[1]])^ sbx8[E[0]];	
	temp2[3] = sbx8[E[7]] ^ sbx1[E[6]] ^ sbx4[E[5]] ^ sbx1[E[4]] ^ sbx1[E[3]] ^ sbx2[E[2]] ^(sbx2[E[1]] ^ sbx8[E[1]])^(sbx1[E[0]] ^ sbx8[E[0]]);	
	temp2[2] =(sbx1[E[7]] ^ sbx8[E[7]])^ sbx8[E[6]] ^ sbx1[E[5]] ^ sbx4[E[4]] ^ sbx1[E[3]] ^ sbx1[E[2]] ^ sbx2[E[1]] ^(sbx2[E[0]] ^ sbx8[E[0]]);	
	temp2[1] =(sbx2[E[7]] ^ sbx8[E[7]])^(sbx1[E[6]] ^ sbx8[E[6]])^ sbx8[E[5]] ^ sbx1[E[4]] ^ sbx4[E[3]] ^ sbx1[E[2]] ^ sbx1[E[1]] ^ sbx2[E[0]];	
	temp2[0] = sbx2[E[7]] ^(sbx2[E[6]] ^ sbx8[E[6]])^(sbx1[E[5]] ^ sbx8[E[5]])^ sbx8[E[4]] ^ sbx1[E[3]] ^ sbx4[E[2]] ^ sbx1[E[1]] ^ sbx1[E[0]];	
	B[0] ^= temp1[0];	B[1] ^= temp1[1];	B[2] ^= temp1[2];	B[3] ^= temp1[3];										
	B[4] ^= temp1[4];	B[5] ^= temp1[5];	B[6] ^= temp1[6];	B[7] ^= temp1[7];										
	SSS3(C,temp1,3,5);																									
	SSS5(D,temp1,7,1);																									
	F[0] ^= temp2[0];	F[1] ^= temp2[1];	F[2] ^= temp2[2];	F[3] ^= temp2[3];										
	F[4] ^= temp2[4];	F[5] ^= temp2[5];	F[6] ^= temp2[6];	F[7] ^= temp2[7];										
	SSS6(G,temp2,3,5);																									
	SSS1(H,temp2,3,5);																									
	}																													


#define COUNT512(){												\
	      state->counter[ 0]++; if(state->counter[ 0]!=0x00);	\
	else {state->counter[ 1]++; if(state->counter[ 1]!=0x00);	\
	else {state->counter[ 2]++; if(state->counter[ 2]!=0x00);	\
	else {state->counter[ 3]++; if(state->counter[ 3]!=0x00);	\
	else {state->counter[ 4]++; if(state->counter[ 4]!=0x00);	\
	else {state->counter[ 5]++; if(state->counter[ 5]!=0x00);	\
	else {state->counter[ 6]++; if(state->counter[ 6]!=0x00);	\
	else {state->counter[ 7]++; if(state->counter[ 7]!=0x00);	\
	else {state->counter[ 8]++; if(state->counter[ 8]!=0x00);	\
	else {state->counter[ 9]++; if(state->counter[ 9]!=0x00);	\
	else {state->counter[10]++; if(state->counter[10]!=0x00);	\
	else {state->counter[11]++; if(state->counter[11]!=0x00);	\
	else {state->counter[12]++; if(state->counter[12]!=0x00);	\
	else {state->counter[13]++; if(state->counter[13]!=0x00);	\
	else {state->counter[14]++; if(state->counter[14]!=0x00);	\
	else {state->counter[15]++; }}}}}}}}}}}}}}}}				\
		
#define COUNT256(){												\
	      state->counter[0]++; if(state->counter[0]!=0x00);		\
	else {state->counter[1]++; if(state->counter[1]!=0x00);		\
	else {state->counter[2]++; if(state->counter[2]!=0x00);		\
	else {state->counter[3]++; if(state->counter[3]!=0x00);		\
	else {state->counter[4]++; if(state->counter[4]!=0x00);		\
	else {state->counter[5]++; if(state->counter[5]!=0x00);		\
	else {state->counter[6]++; if(state->counter[6]!=0x00);		\
	else {state->counter[7]++; }}}}}}}}							\


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : Arirang_Compression256
//
// DESCRIPTION    : Counter addition and compression function of Arirang-224 and ARIRANG-256
//
// PARAMETERS     : state - a structure that holds the hashState information
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////

void Arirang_Compression256(hashState *state)
{
	BYTE	R[32], W[128], temp1[4];
	BYTE	i;

	// Message Scheduling
	for (i = 0; i < 64; i++)
		W[i] = state->block[i];
	for (; i < 128; i++)
		W[i] = 0;

	temp1[0] = (W[ 36] ^ W[ 44] ^ W[ 52] ^ W[ 60] ^ K256[  0]);	temp1[1] = (W[ 37] ^ W[ 45] ^ W[ 53] ^ W[ 61] ^ K256[  1]);
	temp1[2] = (W[ 38] ^ W[ 46] ^ W[ 54] ^ W[ 62] ^ K256[  2]);	temp1[3] = (W[ 39] ^ W[ 47] ^ W[ 55] ^ W[ 63] ^ K256[  3]);
	SS0((W+ 64),temp1,3,5);
	temp1[0] = (W[ 32] ^ W[ 40] ^ W[ 48] ^ W[ 56] ^ K256[  4]);	temp1[1] = (W[ 33] ^ W[ 41] ^ W[ 49] ^ W[ 57] ^ K256[  5]);
	temp1[2] = (W[ 34] ^ W[ 42] ^ W[ 50] ^ W[ 58] ^ K256[  6]);	temp1[3] = (W[ 35] ^ W[ 43] ^ W[ 51] ^ W[ 59] ^ K256[  7]);
	SS1((W+ 68),temp1,5,3);
	temp1[0] = (W[  4] ^ W[ 12] ^ W[ 20] ^ W[ 28] ^ K256[  8]);	temp1[1] = (W[  5] ^ W[ 13] ^ W[ 21] ^ W[ 29] ^ K256[  9]);
	temp1[2] = (W[  6] ^ W[ 14] ^ W[ 22] ^ W[ 30] ^ K256[ 10]);	temp1[3] = (W[  7] ^ W[ 15] ^ W[ 23] ^ W[ 31] ^ K256[ 11]);
	SS2((W+ 72),temp1,5,3);
	temp1[0] = (W[  0] ^ W[  8] ^ W[ 16] ^ W[ 24] ^ K256[ 12]);	temp1[1] = (W[  1] ^ W[  9] ^ W[ 17] ^ W[ 25] ^ K256[ 13]);
	temp1[2] = (W[  2] ^ W[ 10] ^ W[ 18] ^ W[ 26] ^ K256[ 14]);	temp1[3] = (W[  3] ^ W[ 11] ^ W[ 19] ^ W[ 27] ^ K256[ 15]);
	SS3((W+ 76),temp1,1,7);

	temp1[0] = (W[ 56] ^ W[ 16] ^ W[ 40] ^ W[  0] ^ K256[ 16]);	temp1[1] = (W[ 57] ^ W[ 17] ^ W[ 41] ^ W[  1] ^ K256[ 17]);
	temp1[2] = (W[ 58] ^ W[ 18] ^ W[ 42] ^ W[  2] ^ K256[ 18]);	temp1[3] = (W[ 59] ^ W[ 19] ^ W[ 43] ^ W[  3] ^ K256[ 19]);
	SS0((W+ 80),temp1,3,5);
	temp1[0] = (W[ 44] ^ W[  4] ^ W[ 28] ^ W[ 52] ^ K256[ 20]);	temp1[1] = (W[ 45] ^ W[  5] ^ W[ 29] ^ W[ 53] ^ K256[ 21]);
	temp1[2] = (W[ 46] ^ W[  6] ^ W[ 30] ^ W[ 54] ^ K256[ 22]);	temp1[3] = (W[ 47] ^ W[  7] ^ W[ 31] ^ W[ 55] ^ K256[ 23]);
	SS1((W+ 84),temp1,5,3);
	temp1[0] = (W[ 24] ^ W[ 48] ^ W[  8] ^ W[ 32] ^ K256[ 24]);	temp1[1] = (W[ 25] ^ W[ 49] ^ W[  9] ^ W[ 33] ^ K256[ 25]);
	temp1[2] = (W[ 26] ^ W[ 50] ^ W[ 10] ^ W[ 34] ^ K256[ 26]);	temp1[3] = (W[ 27] ^ W[ 51] ^ W[ 11] ^ W[ 35] ^ K256[ 27]);
	SS2((W+ 88),temp1,5,3);
	temp1[0] = (W[ 12] ^ W[ 36] ^ W[ 60] ^ W[ 20] ^ K256[ 28]);	temp1[1] = (W[ 13] ^ W[ 37] ^ W[ 61] ^ W[ 21] ^ K256[ 29]);
	temp1[2] = (W[ 14] ^ W[ 38] ^ W[ 62] ^ W[ 22] ^ K256[ 30]);	temp1[3] = (W[ 15] ^ W[ 39] ^ W[ 63] ^ W[ 23] ^ K256[ 31]);
	SS3((W+ 92),temp1,1,7);

	temp1[0] = (W[ 52] ^ W[ 60] ^ W[  4] ^ W[ 12] ^ K256[ 32]);	temp1[1] = (W[ 53] ^ W[ 61] ^ W[  5] ^ W[ 13] ^ K256[ 33]);
	temp1[2] = (W[ 54] ^ W[ 62] ^ W[  6] ^ W[ 14] ^ K256[ 34]);	temp1[3] = (W[ 55] ^ W[ 63] ^ W[  7] ^ W[ 15] ^ K256[ 35]);
	SS0((W+ 96),temp1,3,5);
	temp1[0] = (W[ 16] ^ W[ 24] ^ W[ 32] ^ W[ 40] ^ K256[ 36]);	temp1[1] = (W[ 17] ^ W[ 25] ^ W[ 33] ^ W[ 41] ^ K256[ 37]);
	temp1[2] = (W[ 18] ^ W[ 26] ^ W[ 34] ^ W[ 42] ^ K256[ 38]);	temp1[3] = (W[ 19] ^ W[ 27] ^ W[ 35] ^ W[ 43] ^ K256[ 39]);
	SS1((W+100),temp1,5,3);
	temp1[0] = (W[ 20] ^ W[ 28] ^ W[ 36] ^ W[ 44] ^ K256[ 40]);	temp1[1] = (W[ 21] ^ W[ 29] ^ W[ 37] ^ W[ 45] ^ K256[ 41]);
	temp1[2] = (W[ 22] ^ W[ 30] ^ W[ 38] ^ W[ 46] ^ K256[ 42]);	temp1[3] = (W[ 23] ^ W[ 31] ^ W[ 39] ^ W[ 47] ^ K256[ 43]);
	SS2((W+104),temp1,5,3);
	temp1[0] = (W[ 48] ^ W[ 56] ^ W[  0] ^ W[  8] ^ K256[ 44]);	temp1[1] = (W[ 49] ^ W[ 57] ^ W[  1] ^ W[  9] ^ K256[ 45]);
	temp1[2] = (W[ 50] ^ W[ 58] ^ W[  2] ^ W[ 10] ^ K256[ 46]);	temp1[3] = (W[ 51] ^ W[ 59] ^ W[  3] ^ W[ 11] ^ K256[ 47]);
	SS3((W+108),temp1,1,7);

	temp1[0] = (W[ 40] ^ W[  0] ^ W[ 24] ^ W[ 48] ^ K256[ 48]);	temp1[1] = (W[ 41] ^ W[  1] ^ W[ 25] ^ W[ 49] ^ K256[ 49]);
	temp1[2] = (W[ 42] ^ W[  2] ^ W[ 26] ^ W[ 50] ^ K256[ 50]);	temp1[3] = (W[ 43] ^ W[  3] ^ W[ 27] ^ W[ 51] ^ K256[ 51]);
	SS0((W+112),temp1,3,5);
	temp1[0] = (W[ 60] ^ W[ 20] ^ W[ 44] ^ W[  4] ^ K256[ 52]);	temp1[1] = (W[ 61] ^ W[ 21] ^ W[ 45] ^ W[  5] ^ K256[ 53]);
	temp1[2] = (W[ 62] ^ W[ 22] ^ W[ 46] ^ W[  6] ^ K256[ 54]);	temp1[3] = (W[ 63] ^ W[ 23] ^ W[ 47] ^ W[  7] ^ K256[ 55]);
	SS1((W+116),temp1,5,3);
	temp1[0] = (W[  8] ^ W[ 32] ^ W[ 56] ^ W[ 16] ^ K256[ 56]);	temp1[1] = (W[  9] ^ W[ 33] ^ W[ 57] ^ W[ 17] ^ K256[ 57]);
	temp1[2] = (W[ 10] ^ W[ 34] ^ W[ 58] ^ W[ 18] ^ K256[ 58]);	temp1[3] = (W[ 11] ^ W[ 35] ^ W[ 59] ^ W[ 19] ^ K256[ 59]);
	SS2((W+120),temp1,5,3);
	temp1[0] = (W[ 28] ^ W[ 52] ^ W[ 12] ^ W[ 36] ^ K256[ 60]);	temp1[1] = (W[ 29] ^ W[ 53] ^ W[ 13] ^ W[ 37] ^ K256[ 61]);
	temp1[2] = (W[ 30] ^ W[ 54] ^ W[ 14] ^ W[ 38] ^ K256[ 62]);	temp1[3] = (W[ 31] ^ W[ 55] ^ W[ 15] ^ W[ 39] ^ K256[ 63]);
	SS3((W+124),temp1,1,7);

	// Counter Addition
	state->workingvar[ 0] ^= state->counter[7];
	state->workingvar[ 1] ^= state->counter[6];
	state->workingvar[ 2] ^= state->counter[5];
	state->workingvar[ 3] ^= state->counter[4];
	state->workingvar[16] ^= state->counter[3];
	state->workingvar[17] ^= state->counter[2];
	state->workingvar[18] ^= state->counter[1];
	state->workingvar[19] ^= state->counter[0];

	// Initialize Register
	R[ 0] = state->workingvar[ 0];	R[ 1] = state->workingvar[ 1];	R[ 2] = state->workingvar[ 2];	R[ 3] = state->workingvar[ 3];
	R[ 4] = state->workingvar[ 4];	R[ 5] = state->workingvar[ 5];	R[ 6] = state->workingvar[ 6];	R[ 7] = state->workingvar[ 7];
	R[ 8] = state->workingvar[ 8];	R[ 9] = state->workingvar[ 9];	R[10] = state->workingvar[10];	R[11] = state->workingvar[11];
	R[12] = state->workingvar[12];	R[13] = state->workingvar[13];	R[14] = state->workingvar[14];	R[15] = state->workingvar[15];
	R[16] = state->workingvar[16];	R[17] = state->workingvar[17];	R[18] = state->workingvar[18];	R[19] = state->workingvar[19];
	R[20] = state->workingvar[20];	R[21] = state->workingvar[21];	R[22] = state->workingvar[22];	R[23] = state->workingvar[23];
	R[24] = state->workingvar[24];	R[25] = state->workingvar[25];	R[26] = state->workingvar[26];	R[27] = state->workingvar[27];
	R[28] = state->workingvar[28];	R[29] = state->workingvar[29];	R[30] = state->workingvar[30];	R[31] = state->workingvar[31];
		
	// 1 round                                                                                                                                 
	Step256((R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (W+ 64),  (W+ 68));
	Step256((R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (W+  0),  (W+  4));
	Step256((R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (W+  8),  (W+ 12));
	Step256((R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (W+ 16),  (W+ 20));
	Step256((R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (W+ 24),  (W+ 28));
	Step256((R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (W+ 72),  (W+ 76));
	Step256((R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (W+ 32),  (W+ 36));
	Step256((R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (W+ 40),  (W+ 44));
	Step256((R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (W+ 48),  (W+ 52));
	Step256((R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (W+ 56),  (W+ 60));
	                                                                                                                                           
	// 2 round                                                                                                                                 
	Step256((R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (W+ 80),  (W+ 84));
	Step256((R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (W+ 12),  (W+ 24));
	Step256((R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (W+ 36),  (W+ 48));
	Step256((R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (W+ 60),  (W+  8));
	Step256((R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (W+ 20),  (W+ 32));
	Step256((R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (W+ 88),  (W+ 92));
	Step256((R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (W+ 44),  (W+ 56));
	Step256((R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (W+  4),  (W+ 16));
	Step256((R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (W+ 28),  (W+ 40));
	Step256((R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (W+ 52),  (W+  0));

	// Feedforward_1
	R[ 0] ^= state->workingvar[16]; R[ 1] ^= state->workingvar[17]; R[ 2] ^= state->workingvar[18]; R[ 3] ^= state->workingvar[19];
	R[ 4] ^= state->workingvar[20]; R[ 5] ^= state->workingvar[21]; R[ 6] ^= state->workingvar[22]; R[ 7] ^= state->workingvar[23];
	R[ 8] ^= state->workingvar[24]; R[ 9] ^= state->workingvar[25]; R[10] ^= state->workingvar[26]; R[11] ^= state->workingvar[27];
	R[12] ^= state->workingvar[28]; R[13] ^= state->workingvar[29]; R[14] ^= state->workingvar[30]; R[15] ^= state->workingvar[31];
	R[16] ^= state->workingvar[0];  R[17] ^= state->workingvar[1];  R[18] ^= state->workingvar[2];  R[19] ^= state->workingvar[ 3];
	R[20] ^= state->workingvar[4];  R[21] ^= state->workingvar[5];  R[22] ^= state->workingvar[6];  R[23] ^= state->workingvar[ 7];
	R[24] ^= state->workingvar[8];  R[25] ^= state->workingvar[9];  R[26] ^= state->workingvar[10]; R[27] ^= state->workingvar[11];
	R[28] ^= state->workingvar[12]; R[29] ^= state->workingvar[13]; R[30] ^= state->workingvar[14]; R[31] ^= state->workingvar[15];
	
	// 3 round                                                                                                                                 
	Step256((R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (W+ 96),  (W+100));
	Step256((R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (W+ 48),  (W+ 20));
	Step256((R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (W+ 56),  (W+ 28));
	Step256((R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (W+  0),  (W+ 36));
	Step256((R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (W+  8),  (W+ 44));
	Step256((R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (W+104),  (W+108));
	Step256((R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (W+ 16),  (W+ 52));
	Step256((R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (W+ 24),  (W+ 60));
	Step256((R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (W+ 32),  (W+  4));
	Step256((R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (W+ 40),  (W+ 12));
	                                                                                                                                           
	                                                                                                                                           
	// 4 round                                                                                                                                 
	Step256((R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (W+112),  (W+116));
	Step256((R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (W+ 28),  (W+  8));
	Step256((R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (W+ 52),  (W+ 32));
	Step256((R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (W+ 12),  (W+ 56));
	Step256((R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (W+ 36),  (W+ 16));
	Step256((R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (W+120),  (W+124));
	Step256((R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (W+ 60),  (W+ 40));
	Step256((R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (W+ 20),  (W+  0));
	Step256((R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (W+ 44),  (W+ 24));
	Step256((R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (W+  4),  (W+ 48));


	// Feedforward_2
	state->workingvar[ 0] ^= R[ 0]; state->workingvar[ 1] ^= R[ 1]; state->workingvar[ 2] ^= R[ 2]; state->workingvar[ 3] ^= R[ 3];
	state->workingvar[ 4] ^= R[ 4]; state->workingvar[ 5] ^= R[ 5]; state->workingvar[ 6] ^= R[ 6]; state->workingvar[ 7] ^= R[ 7];
	state->workingvar[ 8] ^= R[ 8]; state->workingvar[ 9] ^= R[ 9]; state->workingvar[10] ^= R[10]; state->workingvar[11] ^= R[11];
	state->workingvar[12] ^= R[12]; state->workingvar[13] ^= R[13]; state->workingvar[14] ^= R[14]; state->workingvar[15] ^= R[15];
	state->workingvar[16] ^= R[16]; state->workingvar[17] ^= R[17]; state->workingvar[18] ^= R[18]; state->workingvar[19] ^= R[19];
	state->workingvar[20] ^= R[20]; state->workingvar[21] ^= R[21]; state->workingvar[22] ^= R[22]; state->workingvar[23] ^= R[23];
	state->workingvar[24] ^= R[24]; state->workingvar[25] ^= R[25]; state->workingvar[26] ^= R[26]; state->workingvar[27] ^= R[27];
	state->workingvar[28] ^= R[28]; state->workingvar[29] ^= R[29]; state->workingvar[30] ^= R[30]; state->workingvar[31] ^= R[31];

	COUNT256();
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : ARIRANG_Compression512
//
// DESCRIPTION    : Counter Addition and compression function of ARIRANG-384 and ARIRANG-512
//
// PARAMETERS     : state - a structure that holds the hashState information
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////
void Arirang_Compression512(hashState *state)
{
	BYTE	R[64], W[256], temp1[8];
	BYTE i;
	
	// Message Scheduling
	for (i = 0; i < 128; i++)	W[i] = state->block[i];
	memset(W+128,0,128);
									
	temp1[0] = (W[ 72] ^ W[ 88] ^ W[104] ^ W[120] ^ K512[  0]);     temp1[1] = (W[ 73] ^ W[ 89] ^ W[105] ^ W[121] ^ K512[  1]);
	temp1[2] = (W[ 74] ^ W[ 90] ^ W[106] ^ W[122] ^ K512[  2]);     temp1[3] = (W[ 75] ^ W[ 91] ^ W[107] ^ W[123] ^ K512[  3]);
	temp1[4] = (W[ 76] ^ W[ 92] ^ W[108] ^ W[124] ^ K512[  4]);     temp1[5] = (W[ 77] ^ W[ 93] ^ W[109] ^ W[125] ^ K512[  5]);
	temp1[6] = (W[ 78] ^ W[ 94] ^ W[110] ^ W[126] ^ K512[  6]);     temp1[7] = (W[ 79] ^ W[ 95] ^ W[111] ^ W[127] ^ K512[  7]);
	SSS1((W+128),temp1,5,3);
	temp1[0] = (W[ 64] ^ W[ 80] ^ W[ 96] ^ W[112] ^ K512[  8]);     temp1[1] = (W[ 65] ^ W[ 81] ^ W[ 97] ^ W[113] ^ K512[  9]);
	temp1[2] = (W[ 66] ^ W[ 82] ^ W[ 98] ^ W[114] ^ K512[ 10]);     temp1[3] = (W[ 67] ^ W[ 83] ^ W[ 99] ^ W[115] ^ K512[ 11]);
	temp1[4] = (W[ 68] ^ W[ 84] ^ W[100] ^ W[116] ^ K512[ 12]);     temp1[5] = (W[ 69] ^ W[ 85] ^ W[101] ^ W[117] ^ K512[ 13]);
	temp1[6] = (W[ 70] ^ W[ 86] ^ W[102] ^ W[118] ^ K512[ 14]);     temp1[7] = (W[ 71] ^ W[ 87] ^ W[103] ^ W[119] ^ K512[ 15]);
	SSS2((W+136),temp1,1,7);
	temp1[0] = (W[  8] ^ W[ 24] ^ W[ 40] ^ W[ 56] ^ K512[ 16]);     temp1[1] = (W[  9] ^ W[ 25] ^ W[ 41] ^ W[ 57] ^ K512[ 17]);
	temp1[2] = (W[ 10] ^ W[ 26] ^ W[ 42] ^ W[ 58] ^ K512[ 18]);     temp1[3] = (W[ 11] ^ W[ 27] ^ W[ 43] ^ W[ 59] ^ K512[ 19]);
	temp1[4] = (W[ 12] ^ W[ 28] ^ W[ 44] ^ W[ 60] ^ K512[ 20]);     temp1[5] = (W[ 13] ^ W[ 29] ^ W[ 45] ^ W[ 61] ^ K512[ 21]);
	temp1[6] = (W[ 14] ^ W[ 30] ^ W[ 46] ^ W[ 62] ^ K512[ 22]);     temp1[7] = (W[ 15] ^ W[ 31] ^ W[ 47] ^ W[ 63] ^ K512[ 23]);
	SSS4((W+144),temp1,3,5);
	temp1[0] = (W[  0] ^ W[ 16] ^ W[ 32] ^ W[ 48] ^ K512[ 24]);     temp1[1] = (W[  1] ^ W[ 17] ^ W[ 33] ^ W[ 49] ^ K512[ 25]);
	temp1[2] = (W[  2] ^ W[ 18] ^ W[ 34] ^ W[ 50] ^ K512[ 26]);     temp1[3] = (W[  3] ^ W[ 19] ^ W[ 35] ^ W[ 51] ^ K512[ 27]);
	temp1[4] = (W[  4] ^ W[ 20] ^ W[ 36] ^ W[ 52] ^ K512[ 28]);     temp1[5] = (W[  5] ^ W[ 21] ^ W[ 37] ^ W[ 53] ^ K512[ 29]);
	temp1[6] = (W[  6] ^ W[ 22] ^ W[ 38] ^ W[ 54] ^ K512[ 30]);     temp1[7] = (W[  7] ^ W[ 23] ^ W[ 39] ^ W[ 55] ^ K512[ 31]);
	SSS7((W+152),temp1,5,3);
	temp1[0] = (W[112] ^ W[ 32] ^ W[ 80] ^ W[  0] ^ K512[ 32]);     temp1[1] = (W[113] ^ W[ 33] ^ W[ 81] ^ W[  1] ^ K512[ 33]);
	temp1[2] = (W[114] ^ W[ 34] ^ W[ 82] ^ W[  2] ^ K512[ 34]);     temp1[3] = (W[115] ^ W[ 35] ^ W[ 83] ^ W[  3] ^ K512[ 35]);
	temp1[4] = (W[116] ^ W[ 36] ^ W[ 84] ^ W[  4] ^ K512[ 36]);     temp1[5] = (W[117] ^ W[ 37] ^ W[ 85] ^ W[  5] ^ K512[ 37]);
	temp1[6] = (W[118] ^ W[ 38] ^ W[ 86] ^ W[  6] ^ K512[ 38]);     temp1[7] = (W[119] ^ W[ 39] ^ W[ 87] ^ W[  7] ^ K512[ 39]);
	SSS1((W+160),temp1,5,3);
	temp1[0] = (W[ 88] ^ W[  8] ^ W[ 56] ^ W[104] ^ K512[ 40]);     temp1[1] = (W[ 89] ^ W[  9] ^ W[ 57] ^ W[105] ^ K512[ 41]);
	temp1[2] = (W[ 90] ^ W[ 10] ^ W[ 58] ^ W[106] ^ K512[ 42]);     temp1[3] = (W[ 91] ^ W[ 11] ^ W[ 59] ^ W[107] ^ K512[ 43]);
	temp1[4] = (W[ 92] ^ W[ 12] ^ W[ 60] ^ W[108] ^ K512[ 44]);     temp1[5] = (W[ 93] ^ W[ 13] ^ W[ 61] ^ W[109] ^ K512[ 45]);
	temp1[6] = (W[ 94] ^ W[ 14] ^ W[ 62] ^ W[110] ^ K512[ 46]);     temp1[7] = (W[ 95] ^ W[ 15] ^ W[ 63] ^ W[111] ^ K512[ 47]);
	SSS2((W+168),temp1,1,7);
	temp1[0] = (W[ 48] ^ W[ 96] ^ W[ 16] ^ W[ 64] ^ K512[ 48]);     temp1[1] = (W[ 49] ^ W[ 97] ^ W[ 17] ^ W[ 65] ^ K512[ 49]);
	temp1[2] = (W[ 50] ^ W[ 98] ^ W[ 18] ^ W[ 66] ^ K512[ 50]);     temp1[3] = (W[ 51] ^ W[ 99] ^ W[ 19] ^ W[ 67] ^ K512[ 51]);
	temp1[4] = (W[ 52] ^ W[100] ^ W[ 20] ^ W[ 68] ^ K512[ 52]);     temp1[5] = (W[ 53] ^ W[101] ^ W[ 21] ^ W[ 69] ^ K512[ 53]);
	temp1[6] = (W[ 54] ^ W[102] ^ W[ 22] ^ W[ 70] ^ K512[ 54]);     temp1[7] = (W[ 55] ^ W[103] ^ W[ 23] ^ W[ 71] ^ K512[ 55]);
	SSS4((W+176),temp1,3,5);
	temp1[0] = (W[ 24] ^ W[ 72] ^ W[120] ^ W[ 40] ^ K512[ 56]);     temp1[1] = (W[ 25] ^ W[ 73] ^ W[121] ^ W[ 41] ^ K512[ 57]);
	temp1[2] = (W[ 26] ^ W[ 74] ^ W[122] ^ W[ 42] ^ K512[ 58]);     temp1[3] = (W[ 27] ^ W[ 75] ^ W[123] ^ W[ 43] ^ K512[ 59]);
	temp1[4] = (W[ 28] ^ W[ 76] ^ W[124] ^ W[ 44] ^ K512[ 60]);     temp1[5] = (W[ 29] ^ W[ 77] ^ W[125] ^ W[ 45] ^ K512[ 61]);
	temp1[6] = (W[ 30] ^ W[ 78] ^ W[126] ^ W[ 46] ^ K512[ 62]);     temp1[7] = (W[ 31] ^ W[ 79] ^ W[127] ^ W[ 47] ^ K512[ 63]);
	SSS7((W+184),temp1,5,3);
	temp1[0] = (W[104] ^ W[120] ^ W[  8] ^ W[ 24] ^ K512[ 64]);     temp1[1] = (W[105] ^ W[121] ^ W[  9] ^ W[ 25] ^ K512[ 65]);
	temp1[2] = (W[106] ^ W[122] ^ W[ 10] ^ W[ 26] ^ K512[ 66]);     temp1[3] = (W[107] ^ W[123] ^ W[ 11] ^ W[ 27] ^ K512[ 67]);
	temp1[4] = (W[108] ^ W[124] ^ W[ 12] ^ W[ 28] ^ K512[ 68]);     temp1[5] = (W[109] ^ W[125] ^ W[ 13] ^ W[ 29] ^ K512[ 69]);
	temp1[6] = (W[110] ^ W[126] ^ W[ 14] ^ W[ 30] ^ K512[ 70]);     temp1[7] = (W[111] ^ W[127] ^ W[ 15] ^ W[ 31] ^ K512[ 71]);
	SSS1((W+192),temp1,5,3);
	temp1[0] = (W[ 32] ^ W[ 48] ^ W[ 64] ^ W[ 80] ^ K512[ 72]);     temp1[1] = (W[ 33] ^ W[ 49] ^ W[ 65] ^ W[ 81] ^ K512[ 73]);
	temp1[2] = (W[ 34] ^ W[ 50] ^ W[ 66] ^ W[ 82] ^ K512[ 74]);     temp1[3] = (W[ 35] ^ W[ 51] ^ W[ 67] ^ W[ 83] ^ K512[ 75]);
	temp1[4] = (W[ 36] ^ W[ 52] ^ W[ 68] ^ W[ 84] ^ K512[ 76]);     temp1[5] = (W[ 37] ^ W[ 53] ^ W[ 69] ^ W[ 85] ^ K512[ 77]);
	temp1[6] = (W[ 38] ^ W[ 54] ^ W[ 70] ^ W[ 86] ^ K512[ 78]);     temp1[7] = (W[ 39] ^ W[ 55] ^ W[ 71] ^ W[ 87] ^ K512[ 79]);
	SSS2((W+200),temp1,1,7);
	temp1[0] = (W[ 40] ^ W[ 56] ^ W[ 72] ^ W[ 88] ^ K512[ 80]);     temp1[1] = (W[ 41] ^ W[ 57] ^ W[ 73] ^ W[ 89] ^ K512[ 81]);
	temp1[2] = (W[ 42] ^ W[ 58] ^ W[ 74] ^ W[ 90] ^ K512[ 82]);     temp1[3] = (W[ 43] ^ W[ 59] ^ W[ 75] ^ W[ 91] ^ K512[ 83]);
	temp1[4] = (W[ 44] ^ W[ 60] ^ W[ 76] ^ W[ 92] ^ K512[ 84]);     temp1[5] = (W[ 45] ^ W[ 61] ^ W[ 77] ^ W[ 93] ^ K512[ 85]);
	temp1[6] = (W[ 46] ^ W[ 62] ^ W[ 78] ^ W[ 94] ^ K512[ 86]);     temp1[7] = (W[ 47] ^ W[ 63] ^ W[ 79] ^ W[ 95] ^ K512[ 87]);
	SSS4((W+208),temp1,3,5);
	temp1[0] = (W[ 96] ^ W[112] ^ W[  0] ^ W[ 16] ^ K512[ 88]);     temp1[1] = (W[ 97] ^ W[113] ^ W[  1] ^ W[ 17] ^ K512[ 89]);
	temp1[2] = (W[ 98] ^ W[114] ^ W[  2] ^ W[ 18] ^ K512[ 90]);     temp1[3] = (W[ 99] ^ W[115] ^ W[  3] ^ W[ 19] ^ K512[ 91]);
	temp1[4] = (W[100] ^ W[116] ^ W[  4] ^ W[ 20] ^ K512[ 92]);     temp1[5] = (W[101] ^ W[117] ^ W[  5] ^ W[ 21] ^ K512[ 93]);
	temp1[6] = (W[102] ^ W[118] ^ W[  6] ^ W[ 22] ^ K512[ 94]);     temp1[7] = (W[103] ^ W[119] ^ W[  7] ^ W[ 23] ^ K512[ 95]);
	SSS7((W+216),temp1,5,3);
	temp1[0] = (W[ 80] ^ W[  0] ^ W[ 48] ^ W[ 96] ^ K512[ 96]);     temp1[1] = (W[ 81] ^ W[  1] ^ W[ 49] ^ W[ 97] ^ K512[ 97]);
	temp1[2] = (W[ 82] ^ W[  2] ^ W[ 50] ^ W[ 98] ^ K512[ 98]);     temp1[3] = (W[ 83] ^ W[  3] ^ W[ 51] ^ W[ 99] ^ K512[ 99]);
	temp1[4] = (W[ 84] ^ W[  4] ^ W[ 52] ^ W[100] ^ K512[100]);     temp1[5] = (W[ 85] ^ W[  5] ^ W[ 53] ^ W[101] ^ K512[101]);
	temp1[6] = (W[ 86] ^ W[  6] ^ W[ 54] ^ W[102] ^ K512[102]);     temp1[7] = (W[ 87] ^ W[  7] ^ W[ 55] ^ W[103] ^ K512[103]);
	SSS1((W+224),temp1,5,3);
	temp1[0] = (W[120] ^ W[ 40] ^ W[ 88] ^ W[  8] ^ K512[104]);     temp1[1] = (W[121] ^ W[ 41] ^ W[ 89] ^ W[  9] ^ K512[105]);
	temp1[2] = (W[122] ^ W[ 42] ^ W[ 90] ^ W[ 10] ^ K512[106]);     temp1[3] = (W[123] ^ W[ 43] ^ W[ 91] ^ W[ 11] ^ K512[107]);
	temp1[4] = (W[124] ^ W[ 44] ^ W[ 92] ^ W[ 12] ^ K512[108]);     temp1[5] = (W[125] ^ W[ 45] ^ W[ 93] ^ W[ 13] ^ K512[109]);
	temp1[6] = (W[126] ^ W[ 46] ^ W[ 94] ^ W[ 14] ^ K512[110]);     temp1[7] = (W[127] ^ W[ 47] ^ W[ 95] ^ W[ 15] ^ K512[111]);
	SSS2((W+232),temp1,1,7);
	temp1[0] = (W[ 16] ^ W[ 64] ^ W[112] ^ W[ 32] ^ K512[112]);     temp1[1] = (W[ 17] ^ W[ 65] ^ W[113] ^ W[ 33] ^ K512[113]);
	temp1[2] = (W[ 18] ^ W[ 66] ^ W[114] ^ W[ 34] ^ K512[114]);     temp1[3] = (W[ 19] ^ W[ 67] ^ W[115] ^ W[ 35] ^ K512[115]);
	temp1[4] = (W[ 20] ^ W[ 68] ^ W[116] ^ W[ 36] ^ K512[116]);     temp1[5] = (W[ 21] ^ W[ 69] ^ W[117] ^ W[ 37] ^ K512[117]);
	temp1[6] = (W[ 22] ^ W[ 70] ^ W[118] ^ W[ 38] ^ K512[118]);     temp1[7] = (W[ 23] ^ W[ 71] ^ W[119] ^ W[ 39] ^ K512[119]);
	SSS4((W+240),temp1,3,5);
	temp1[0] = (W[ 56] ^ W[104] ^ W[ 24] ^ W[ 72] ^ K512[120]);     temp1[1] = (W[ 57] ^ W[105] ^ W[ 25] ^ W[ 73] ^ K512[121]);
	temp1[2] = (W[ 58] ^ W[106] ^ W[ 26] ^ W[ 74] ^ K512[122]);     temp1[3] = (W[ 59] ^ W[107] ^ W[ 27] ^ W[ 75] ^ K512[123]);
	temp1[4] = (W[ 60] ^ W[108] ^ W[ 28] ^ W[ 76] ^ K512[124]);     temp1[5] = (W[ 61] ^ W[109] ^ W[ 29] ^ W[ 77] ^ K512[125]);
	temp1[6] = (W[ 62] ^ W[110] ^ W[ 30] ^ W[ 78] ^ K512[126]);     temp1[7] = (W[ 63] ^ W[111] ^ W[ 31] ^ W[ 79] ^ K512[127]);
	SSS7((W+248),temp1,5,3);																														

	// Counter Addition
	state->workingvar[ 0] ^= state->counter[15];
	state->workingvar[ 1] ^= state->counter[14];
	state->workingvar[ 2] ^= state->counter[13];
	state->workingvar[ 3] ^= state->counter[12];
	state->workingvar[ 4] ^= state->counter[11];
	state->workingvar[ 5] ^= state->counter[10];
	state->workingvar[ 6] ^= state->counter[ 9];
	state->workingvar[ 7] ^= state->counter[ 8];
	state->workingvar[32] ^= state->counter[ 7];
	state->workingvar[33] ^= state->counter[ 6];
	state->workingvar[34] ^= state->counter[ 5];
	state->workingvar[35] ^= state->counter[ 4];
	state->workingvar[36] ^= state->counter[ 3];
	state->workingvar[37] ^= state->counter[ 2];
	state->workingvar[38] ^= state->counter[ 1];
	state->workingvar[39] ^= state->counter[ 0];

	// Initialize Register
	R[ 0] = state->workingvar[ 0];	R[ 1] = state->workingvar[ 1];	R[ 2] = state->workingvar[ 2];	R[ 3] = state->workingvar[ 3];
	R[ 4] = state->workingvar[ 4];	R[ 5] = state->workingvar[ 5];	R[ 6] = state->workingvar[ 6];	R[ 7] = state->workingvar[ 7];
	R[ 8] = state->workingvar[ 8];	R[ 9] = state->workingvar[ 9];	R[10] = state->workingvar[10];	R[11] = state->workingvar[11];
	R[12] = state->workingvar[12];	R[13] = state->workingvar[13];	R[14] = state->workingvar[14];	R[15] = state->workingvar[15];
	R[16] = state->workingvar[16];	R[17] = state->workingvar[17];	R[18] = state->workingvar[18];	R[19] = state->workingvar[19];
	R[20] = state->workingvar[20];	R[21] = state->workingvar[21];	R[22] = state->workingvar[22];	R[23] = state->workingvar[23];
	R[24] = state->workingvar[24];	R[25] = state->workingvar[25];	R[26] = state->workingvar[26];	R[27] = state->workingvar[27];
	R[28] = state->workingvar[28];	R[29] = state->workingvar[29];	R[30] = state->workingvar[30];	R[31] = state->workingvar[31];
	R[32] = state->workingvar[32];	R[33] = state->workingvar[33];	R[34] = state->workingvar[34];	R[35] = state->workingvar[35];
	R[36] = state->workingvar[36];	R[37] = state->workingvar[37];	R[38] = state->workingvar[38];	R[39] = state->workingvar[39];
	R[40] = state->workingvar[40];	R[41] = state->workingvar[41];	R[42] = state->workingvar[42];	R[43] = state->workingvar[43];
	R[44] = state->workingvar[44];	R[45] = state->workingvar[45];	R[46] = state->workingvar[46];	R[47] = state->workingvar[47];
	R[48] = state->workingvar[48];	R[49] = state->workingvar[49];	R[50] = state->workingvar[50];	R[51] = state->workingvar[51];
	R[52] = state->workingvar[52];	R[53] = state->workingvar[53];	R[54] = state->workingvar[54];	R[55] = state->workingvar[55];
	R[56] = state->workingvar[56];	R[57] = state->workingvar[57];	R[58] = state->workingvar[58];	R[59] = state->workingvar[59];
	R[60] = state->workingvar[60];	R[61] = state->workingvar[61];	R[62] = state->workingvar[62];	R[63] = state->workingvar[63];
	
	// 1 Round                                                                                      
	Step512((R+ 0), (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56), (W+128), (W+136));
	Step512((R+56), (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48), (W+  0), (W+  8));
	Step512((R+48), (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40), (W+ 16), (W+ 24));
	Step512((R+40), (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32), (W+ 32), (W+ 40));
	Step512((R+32), (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24), (W+ 48), (W+ 56));
	Step512((R+24), (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16), (W+144), (W+152));
	Step512((R+16), (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8), (W+ 64), (W+ 72));
	Step512((R+ 8), (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0), (W+ 80), (W+ 88));
	Step512((R+ 0), (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56), (W+ 96), (W+104));
	Step512((R+56), (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48), (W+112), (W+120));
                                                                                                
	// 2 round                                                                                    
	Step512((R+48), (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40), (W+160), (W+168));
	Step512((R+40), (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32), (W+ 24), (W+ 48));
	Step512((R+32), (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24), (W+ 72), (W+ 96));
	Step512((R+24), (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16), (W+120), (W+ 16));
	Step512((R+16), (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8), (W+ 40), (W+ 64));
	Step512((R+ 8), (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0), (W+176), (W+184));
	Step512((R+ 0), (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56), (W+ 88), (W+112));
	Step512((R+56), (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48), (W+  8), (W+ 32));
	Step512((R+48), (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40), (W+ 56), (W+ 80));
	Step512((R+40), (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32), (W+104), (W+  0));



	// Feedforward_1
	R[ 0] ^= state->workingvar[32]; R[ 1] ^= state->workingvar[33];	R[ 2] ^= state->workingvar[34]; R[ 3] ^= state->workingvar[35];
	R[ 4] ^= state->workingvar[36]; R[ 5] ^= state->workingvar[37];	R[ 6] ^= state->workingvar[38]; R[ 7] ^= state->workingvar[39];
	R[ 8] ^= state->workingvar[40]; R[ 9] ^= state->workingvar[41];	R[10] ^= state->workingvar[42]; R[11] ^= state->workingvar[43];
	R[12] ^= state->workingvar[44]; R[13] ^= state->workingvar[45];	R[14] ^= state->workingvar[46]; R[15] ^= state->workingvar[47];
	R[16] ^= state->workingvar[48]; R[17] ^= state->workingvar[49];	R[18] ^= state->workingvar[50]; R[19] ^= state->workingvar[51];
	R[20] ^= state->workingvar[52]; R[21] ^= state->workingvar[53];	R[22] ^= state->workingvar[54]; R[23] ^= state->workingvar[55];
	R[24] ^= state->workingvar[56]; R[25] ^= state->workingvar[57];	R[26] ^= state->workingvar[58]; R[27] ^= state->workingvar[59];
	R[28] ^= state->workingvar[60]; R[29] ^= state->workingvar[61];	R[30] ^= state->workingvar[62]; R[31] ^= state->workingvar[63];
	R[32] ^= state->workingvar[ 0]; R[33] ^= state->workingvar[ 1];	R[34] ^= state->workingvar[ 2]; R[35] ^= state->workingvar[ 3];
	R[36] ^= state->workingvar[ 4]; R[37] ^= state->workingvar[ 5];	R[38] ^= state->workingvar[ 6]; R[39] ^= state->workingvar[ 7];
	R[40] ^= state->workingvar[ 8]; R[41] ^= state->workingvar[ 9];	R[42] ^= state->workingvar[10]; R[43] ^= state->workingvar[11];
	R[44] ^= state->workingvar[12]; R[45] ^= state->workingvar[13];	R[46] ^= state->workingvar[14]; R[47] ^= state->workingvar[15];
	R[48] ^= state->workingvar[16]; R[49] ^= state->workingvar[17];	R[50] ^= state->workingvar[18]; R[51] ^= state->workingvar[19];
	R[52] ^= state->workingvar[20]; R[53] ^= state->workingvar[21];	R[54] ^= state->workingvar[22]; R[55] ^= state->workingvar[23];
	R[56] ^= state->workingvar[24]; R[57] ^= state->workingvar[25];	R[58] ^= state->workingvar[26]; R[59] ^= state->workingvar[27];
	R[60] ^= state->workingvar[28]; R[61] ^= state->workingvar[29];	R[62] ^= state->workingvar[30]; R[63] ^= state->workingvar[31];

	                                                                                   
	// 3 Round                                                                                    
	Step512((R+32), (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24), (W+192), (W+200));
	Step512((R+24), (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16), (W+ 96), (W+ 40));
	Step512((R+16), (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8), (W+112), (W+ 56));
	Step512((R+ 8), (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0), (W+  0), (W+ 72));
	Step512((R+ 0), (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56), (W+ 16), (W+ 88));
	Step512((R+56), (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48), (W+208), (W+216));
	Step512((R+48), (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40), (W+ 32), (W+104));
	Step512((R+40), (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32), (W+ 48), (W+120));
	Step512((R+32), (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24), (W+ 64), (W+  8));
	Step512((R+24), (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16), (W+ 80), (W+ 24));
	                                                                                              
	// 4 round                                                                                    
	Step512((R+16), (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8), (W+224), (W+232));
	Step512((R+ 8), (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0), (W+ 56), (W+ 16));
	Step512((R+ 0), (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56), (W+104), (W+ 64));
	Step512((R+56), (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48), (W+ 24), (W+112));
	Step512((R+48), (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40), (W+ 72), (W+ 32));
	Step512((R+40), (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32), (W+240), (W+248));
	Step512((R+32), (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24), (W+120), (W+ 80));
	Step512((R+24), (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16), (W+ 40), (W+  0));
	Step512((R+16), (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8), (W+ 88), (W+ 48));
	Step512((R+ 8), (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0), (W+  8), (W+ 96));
	

	// Feedforward_2
	state->workingvar[ 0] ^= R[ 0]; state->workingvar[ 1] ^= R[ 1];	state->workingvar[ 2] ^= R[ 2]; state->workingvar[ 3] ^= R[ 3];
	state->workingvar[ 4] ^= R[ 4]; state->workingvar[ 5] ^= R[ 5];	state->workingvar[ 6] ^= R[ 6]; state->workingvar[ 7] ^= R[ 7];
	state->workingvar[ 8] ^= R[ 8]; state->workingvar[ 9] ^= R[ 9];	state->workingvar[10] ^= R[10]; state->workingvar[11] ^= R[11];
	state->workingvar[12] ^= R[12]; state->workingvar[13] ^= R[13];	state->workingvar[14] ^= R[14]; state->workingvar[15] ^= R[15];
	state->workingvar[16] ^= R[16]; state->workingvar[17] ^= R[17];	state->workingvar[18] ^= R[18]; state->workingvar[19] ^= R[19];
	state->workingvar[20] ^= R[20]; state->workingvar[21] ^= R[21];	state->workingvar[22] ^= R[22]; state->workingvar[23] ^= R[23];
	state->workingvar[24] ^= R[24]; state->workingvar[25] ^= R[25];	state->workingvar[26] ^= R[26]; state->workingvar[27] ^= R[27];
	state->workingvar[28] ^= R[28]; state->workingvar[29] ^= R[29];	state->workingvar[30] ^= R[30]; state->workingvar[31] ^= R[31];
	state->workingvar[32] ^= R[32]; state->workingvar[33] ^= R[33];	state->workingvar[34] ^= R[34]; state->workingvar[35] ^= R[35];
	state->workingvar[36] ^= R[36]; state->workingvar[37] ^= R[37];	state->workingvar[38] ^= R[38]; state->workingvar[39] ^= R[39];
	state->workingvar[40] ^= R[40]; state->workingvar[41] ^= R[41];	state->workingvar[42] ^= R[42]; state->workingvar[43] ^= R[43];
	state->workingvar[44] ^= R[44]; state->workingvar[45] ^= R[45];	state->workingvar[46] ^= R[46]; state->workingvar[47] ^= R[47];
	state->workingvar[48] ^= R[48]; state->workingvar[49] ^= R[49];	state->workingvar[50] ^= R[50]; state->workingvar[51] ^= R[51];
	state->workingvar[52] ^= R[52]; state->workingvar[53] ^= R[53];	state->workingvar[54] ^= R[54]; state->workingvar[55] ^= R[55];
	state->workingvar[56] ^= R[56]; state->workingvar[57] ^= R[57];	state->workingvar[58] ^= R[58]; state->workingvar[59] ^= R[59];
	state->workingvar[60] ^= R[60]; state->workingvar[61] ^= R[61]; state->workingvar[62] ^= R[62]; state->workingvar[63] ^= R[63];
	
	COUNT512();														
}

